import {
  StateResource,
  createEmptyStateResource,
  createStateResource,
} from '@alfa-client/tech-shared';
import { Injectable, WritableSignal, signal } from '@angular/core';
import { KeycloakEvent, KeycloakEventType, KeycloakService } from 'keycloak-angular';
import { KeycloakProfile } from 'keycloak-js';
import { UserProfileResource } from 'libs/user-profile-shared/src/lib/user-profile.model';
import { getUserNameInitials } from 'libs/user-profile-shared/src/lib/user-profile.util';

@Injectable({ providedIn: 'root' })
export class AuthService {
  currentUserStateResource: WritableSignal<StateResource<UserProfileResource>> = signal<
    StateResource<UserProfileResource>
  >(createEmptyStateResource());

  public isLoggedIn = false;

  constructor(private keycloak: KeycloakService) {
    this.listenToKeycloakEvent();
  }

  listenToKeycloakEvent(): void {
    this.keycloak.keycloakEvents$.subscribe((event: KeycloakEvent) => this.handleEvent(event));
  }

  async handleEvent(event: KeycloakEvent): Promise<void> {
    if (event.type === KeycloakEventType.OnTokenExpired) {
      await this.keycloak
        .updateToken(10)
        .then((refreshed) => this.handleTokenExpiredEvent(refreshed));
    }
  }

  async handleTokenExpiredEvent(refreshed: boolean): Promise<void> {
    if (refreshed) {
      await this.keycloak.getToken().then((token) => this.storeAccessToken(token));
    }
  }

  public async login(): Promise<void> {
    this.isLoggedIn = await this.keycloak.isLoggedIn();
    if (this.isLoggedIn) {
      const token: string = await this.keycloak.getToken();
      this.storeAccessToken(token);
      await this.setCurrentUser();
    } else {
      this.keycloak.login({ scope: 'offline_access' });
    }
  }

  storeAccessToken(token: string): void {
    sessionStorage.setItem('access_token', token);
  }

  async setCurrentUser(): Promise<void> {
    const profile: KeycloakProfile = await this.keycloak.loadUserProfile();

    this.currentUserStateResource.set(createStateResource(this.buildUserProfileResource(profile)));
  }

  private buildUserProfileResource(profile: KeycloakProfile): UserProfileResource {
    return <any>{
      firstName: profile.firstName,
      lastName: profile.lastName,
    };
  }

  public logout(): void {
    this.keycloak.logout();
  }

  public getCurrentUserInitials(): string {
    return getUserNameInitials(this.currentUserStateResource().resource);
  }
}
