/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { encodeUrlForEmbedding, isNotNull } from '@alfa-client/tech-shared';
import { Resource, ResourceUri, getEmbeddedResource, getUrl } from '@ngxp/rest';
import { isNil } from 'lodash-es';
import { ApiError } from '../tech.model';

export interface ListResource extends Resource {
  _embedded;
}

export interface StateResource<T> {
  resource: T;
  reload: boolean;
  loading: boolean;
  loaded: boolean;
  error?: ApiError;
}

export function createEmptyStateResource<T>(loading: boolean = false): StateResource<T> {
  return createStateResource(null, loading);
}

export function createStateResource<T>(resource: T, loading: boolean = false): StateResource<T> {
  return { loading, loaded: !isNil(resource), reload: false, resource };
}

export function createErrorStateResource<T>(error: ApiError): StateResource<any> {
  return { ...createEmptyStateResource<T>(), error, loaded: true };
}

export function createAnyErrorStateResource<T>(error: any): StateResource<any> {
  return { ...createEmptyStateResource<T>(), error, loaded: true };
}

export function doIfLoadingRequired(
  stateResource: StateResource<any>,
  runable: () => void,
): boolean {
  if (isLoadingRequired(stateResource)) {
    runable();
    return true;
  }
  return false;
}

export function isLoadingRequired(stateResource: StateResource<any>): boolean {
  return !stateResource.loading && (!stateResource.loaded || stateResource.reload);
}

export function isLoaded(stateResource: StateResource<Resource>): boolean {
  return !stateResource.loading && !stateResource.reload && isNotNull(stateResource.resource);
}

export function hasError(stateResource: StateResource<any>): boolean {
  return !isNil(stateResource.error);
}

export function toResourceUri(resource: Resource, linkRel?: string): string {
  const url: ResourceUri = isNil(linkRel) ? getUrl(resource) : getUrl(resource, linkRel);
  return encodeUrlForEmbedding(url);
}

export function doOnValidStateResource(
  stateResource: StateResource<Resource>,
  actionOnValid: () => void,
): void {
  if (isValidStateResouce(stateResource)) actionOnValid();
}

export function isValidStateResouce(stateResource: StateResource<Resource>): boolean {
  return stateResource.loaded && isNotNull(stateResource.resource);
}

export function getEmbeddedResources<T>(
  stateResource: StateResource<Resource>,
  linkRel: string,
): T[] {
  if (isNil(stateResource) || isNil(stateResource.resource)) {
    return [];
  }
  return getEmbeddedResource<T[]>(stateResource.resource, linkRel);
}
