/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { BinaryFileListResource, BinaryFileService } from '@alfa-client/binary-file-shared';
import {
  Command,
  CommandOrder,
  CommandResource,
  CommandService,
  CreateCommand,
  isDone,
} from '@alfa-client/command-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import {
  StateResource,
  createEmptyStateResource,
  createStateResource,
  decodeUrlFromEmbedding,
  doIfLoadingRequired,
  hasError,
  isNotNull,
  isNotUndefined,
  replacePlaceholder,
} from '@alfa-client/tech-shared';
import { SnackBarService } from '@alfa-client/ui';
import { VorgangHeaderLinkRel, VorgangResource, VorgangService } from '@alfa-client/vorgang-shared';
import { Injectable, OnDestroy } from '@angular/core';
import { Params } from '@angular/router';
import { ResourceUri, getUrl, hasLink } from '@ngxp/rest';
import { isEqual, isNil, isNull, isUndefined } from 'lodash-es';
import { BehaviorSubject, Observable, Subscription, combineLatest } from 'rxjs';
import { filter, map, startWith, tap } from 'rxjs/operators';
import { WiedervorlageLinkRel, WiedervorlageListLinkRel } from './wiedervorlage.linkrel';
import { WiedervorlageMessages } from './wiedervorlage.message';
import {
  BinaryFileListByWiedervorlageUri,
  CreateWiedervorlageCommand,
  Wiedervorlage,
  WiedervorlageListResource,
  WiedervorlageResource,
} from './wiedervorlage.model';
import { WiedervorlageRepository } from './wiedervorlage.repository';
import { WiedervorlageRoutes } from './wiedervorlage.route';
import { createErledigenCommand, createWiedereroeffnenCommand } from './wiedervorlage.util';

@Injectable({ providedIn: 'root' })
export class WiedervorlageService implements OnDestroy {
  static encodedWiedervorlageUriParam: string = 'wiedervorlageUrl';

  readonly wiedervorlageList$: BehaviorSubject<StateResource<WiedervorlageListResource>> =
    new BehaviorSubject<StateResource<WiedervorlageListResource>>(
      createEmptyStateResource<WiedervorlageListResource>(),
    );
  readonly wiedervorlage$: BehaviorSubject<StateResource<WiedervorlageResource>> =
    new BehaviorSubject<StateResource<WiedervorlageResource>>(
      createEmptyStateResource<WiedervorlageResource>(),
    );

  readonly submitInProgress$: BehaviorSubject<StateResource<CommandResource>> = new BehaviorSubject<
    StateResource<CommandResource>
  >(createEmptyStateResource<CommandResource>());

  readonly erledigenCommand$: BehaviorSubject<StateResource<CommandResource>> = new BehaviorSubject<
    StateResource<CommandResource>
  >(createEmptyStateResource<CommandResource>());
  readonly wiedereroeffnenCommand$: BehaviorSubject<StateResource<CommandResource>> =
    new BehaviorSubject<StateResource<CommandResource>>(
      createEmptyStateResource<CommandResource>(),
    );

  attachmentListByWiedervorlage = <BinaryFileListByWiedervorlageUri>{};

  private subscription: Subscription;
  private attachmentSubscription: Subscription;

  constructor(
    private repository: WiedervorlageRepository,
    private navigationService: NavigationService,
    private commandService: CommandService,
    private snackbarService: SnackBarService,
    private binaryFileService: BinaryFileService,
    private vorgangService: VorgangService,
  ) {
    this.listenToNavigation();
  }

  public getWiedervorlageList(): Observable<StateResource<WiedervorlageListResource>> {
    return combineLatest([
      this.wiedervorlageList$.asObservable(),
      this.vorgangService.getVorgangWithEingang(),
    ]).pipe(
      tap(([wiedervorlageList, vorgangResource]) => {
        if (isNotNull(vorgangResource.resource)) {
          doIfLoadingRequired(wiedervorlageList, () =>
            this.loadWiedervorlagenByVorgang(vorgangResource.resource),
          );
        }
      }),
      map(([wiedervorlageList]) => wiedervorlageList),
      startWith(createEmptyStateResource<WiedervorlageListResource>(true)),
    );
  }

  public getWiedervorlageListByGivenVorgang(
    vorgangResource: VorgangResource,
  ): Observable<StateResource<WiedervorlageListResource>> {
    doIfLoadingRequired(this.wiedervorlageList$.value, () =>
      this.loadWiedervorlagenByVorgang(vorgangResource),
    );
    return this.wiedervorlageList$.asObservable();
  }

  loadWiedervorlagenByVorgang(vorgangResource: VorgangResource): void {
    if (hasLink(vorgangResource, VorgangHeaderLinkRel.WIEDERVORLAGEN)) {
      const subscription: Subscription = this.repository
        .getWiedervorlageList(vorgangResource)
        .subscribe((wiedervorlagenList) => {
          if (!isNull(wiedervorlagenList)) {
            this.setWiedervorlagenList(wiedervorlagenList);
            subscription.unsubscribe();
          }
        });
    } else {
      this.clearWiedervorlagenList();
    }
  }

  public loadWiedervorlagenByVorgangStrict(
    vorgangResource: VorgangResource,
  ): Observable<StateResource<WiedervorlageListResource>> {
    return this.repository.getWiedervorlageList(vorgangResource).pipe(
      map((wiedervorlagenList) => createStateResource(wiedervorlagenList)),
      startWith(createEmptyStateResource<WiedervorlageListResource>(true)),
    );
  }

  setWiedervorlageListLoading(): void {
    this.wiedervorlageList$.next({ ...this.wiedervorlageList$.value, loading: true });
  }

  setWiedervorlagenList(wiedervorlagenList: WiedervorlageListResource): void {
    this.wiedervorlageList$.next(createStateResource(wiedervorlagenList));
  }

  public getWiedervorlage(): Observable<StateResource<WiedervorlageResource>> {
    return this.wiedervorlage$.asObservable().pipe(
      tap((wiedervorlage) => doIfLoadingRequired(wiedervorlage, () => this.updateWiedervorlage())),
      startWith(createEmptyStateResource<WiedervorlageResource>(true)),
    );
  }

  private updateWiedervorlage(): void {
    this.setWiedervorlageLoading();

    this.loadAndSetWiedervorlageByUrl(this.getWiedervorlageUri());
  }

  private loadAndSetWiedervorlageByUrl(url: ResourceUri): void {
    this.setWiedervorlageLoading();

    const subscription: Subscription = this.repository
      .getWiedervorlage(url)
      .subscribe((wiedervorlage) => {
        this.wiedervorlage$.next(createStateResource(wiedervorlage));
        subscription.unsubscribe();
      });
  }

  public isNewWiedervorlage(): boolean {
    return isEqual(
      this.navigationService.getParam(WiedervorlageService.encodedWiedervorlageUriParam),
      WiedervorlageRoutes.NEW,
    );
  }

  private listenToNavigation(): void {
    this.unsubscribe();
    this.subscription = this.navigationService
      .urlChanged()
      .subscribe((params) => this.onNavigation(params));
  }

  onNavigation(params: Params): void {
    if (this.navigateToVorgangDetailPage(params)) {
      this.setWiedervorlageListReload();
      this.forceWiedervorlageReload();
      this.submitInProgress$.next(createEmptyStateResource<CommandResource>());
      this.clearAttachmentList();
    }
    if (NavigationService.isVorgangListPage(params)) {
      this.clearWiedervorlagenList();
      this.submitInProgress$.next(createEmptyStateResource<CommandResource>());
      this.clearAttachmentList();
    }
  }

  clearAttachmentList(): void {
    this.attachmentListByWiedervorlage = <BinaryFileListByWiedervorlageUri>{};
  }

  clearWiedervorlagenList(): void {
    this.wiedervorlageList$.next(createEmptyStateResource<WiedervorlageListResource>());
  }

  private forceWiedervorlageReload(): void {
    this.wiedervorlage$.next(createEmptyStateResource());
  }

  private navigateToVorgangDetailPage(params: Params): boolean {
    return isUndefined(
      params[WiedervorlageService.encodedWiedervorlageUriParam] &&
        isNotUndefined(params[VorgangService.VORGANG_WITH_EINGANG_URL]),
    );
  }

  private getWiedervorlageUri(): ResourceUri {
    return this.navigationService.getDecodedParam(
      WiedervorlageService.encodedWiedervorlageUriParam,
    );
  }

  setWiedervorlageLoading(): void {
    this.wiedervorlage$.next({ ...this.wiedervorlage$.value, loading: true });
  }

  public createWiedervorlage(
    wiedervorlage: Wiedervorlage,
  ): Observable<StateResource<CommandResource>> {
    this.setSubmitInProgressLoading();

    return this.createCreateWiedervorlageCommand(
      this.wiedervorlageList$.value.resource,
      wiedervorlage,
    ).pipe(
      filter((commandStateResource) => !commandStateResource.loading),
      tap((commandStateResource) =>
        this.proceedAfterReceiveCommand(
          commandStateResource,
          this.buildMessage(WiedervorlageMessages.CREATED, wiedervorlage),
        ),
      ),
      startWith(createEmptyStateResource<CommandResource>(true)),
    );
  }

  createCreateWiedervorlageCommand(
    wiedervorlageListResource: WiedervorlageListResource,
    wiedervorlage: Wiedervorlage,
  ): Observable<StateResource<CommandResource>> {
    return this.commandService.createCommand(
      wiedervorlageListResource,
      WiedervorlageListLinkRel.CREATE_WIEDERVORLAGE,
      this.buildCreateWiedervorlageCommand(wiedervorlage),
    );
  }

  private buildCreateWiedervorlageCommand(wiedervorlage: Wiedervorlage): Command {
    return <any>{ order: CommandOrder.CREATE_WIEDERVORLAGE, wiedervorlage, body: null };
  }

  public saveWiedervorlage(
    wiedervorlageResource: WiedervorlageResource,
    wiedervorlage: Wiedervorlage,
  ): Observable<StateResource<CommandResource>> {
    this.setSubmitInProgressLoading();

    return this.createSaveWiedervorlageCommand(wiedervorlageResource, wiedervorlage).pipe(
      filter((commandStateResource) => !commandStateResource.loading),
      tap((commandStateResource) =>
        this.proceedAfterReceiveCommand(
          commandStateResource,
          this.buildMessage(WiedervorlageMessages.SAVED, wiedervorlage),
        ),
      ),
      startWith(createEmptyStateResource<CommandResource>(true)),
    );
  }

  private setSubmitInProgressLoading(): void {
    this.submitInProgress$.next({ ...this.submitInProgress$.value, loading: true });
  }

  createSaveWiedervorlageCommand(
    wiedervorlageResource: WiedervorlageResource,
    wiedervorlage: Wiedervorlage,
  ): Observable<StateResource<CommandResource>> {
    return this.commandService.createCommand(
      wiedervorlageResource,
      WiedervorlageLinkRel.EDIT,
      this.buildEditWiedervorlageCommand(wiedervorlage),
    );
  }

  private buildEditWiedervorlageCommand(wiedervorlage: Wiedervorlage): CreateWiedervorlageCommand {
    return { order: CommandOrder.EDIT_WIEDERVORLAGE, wiedervorlage, body: null };
  }

  proceedAfterReceiveCommand(
    commandStateResource: StateResource<CommandResource>,
    message: string,
  ): void {
    this.submitInProgress$.next({
      ...this.submitInProgress$.value,
      resource: commandStateResource.resource,
    });
    if (isDone(commandStateResource.resource)) {
      this.submitInProgress$.next(createStateResource(commandStateResource.resource));
      this.snackbarService.show(commandStateResource.resource, message);
      this.setWiedervorlageListReload();
    } else if (hasError(commandStateResource)) {
      this.submitInProgress$.next(createStateResource(commandStateResource.resource));
    }
  }

  setWiedervorlageListReload(): void {
    this.wiedervorlageList$.next({ ...this.wiedervorlageList$.value, reload: true });
  }

  public erledigen(): Observable<StateResource<CommandResource>> {
    return this.doAction(
      this.erledigenCommand$,
      WiedervorlageLinkRel.ERLEDIGEN,
      createErledigenCommand,
      WiedervorlageMessages.ERLEDIGT,
    );
  }

  public wiedereroeffnen(): Observable<StateResource<CommandResource>> {
    return this.doAction(
      this.wiedereroeffnenCommand$,
      WiedervorlageLinkRel.WIEDEREROEFFNEN,
      createWiedereroeffnenCommand,
      WiedervorlageMessages.WIEDEREROEFFNET,
    );
  }

  private doAction(
    commandStateSubj: BehaviorSubject<StateResource<CommandResource>>,
    linkRel: string,
    commandProvider: () => CreateCommand,
    snackBarMessage: string,
  ): Observable<StateResource<any>> {
    commandStateSubj.next(createEmptyStateResource(true));

    const commandSubscription: Subscription = this.commandService
      .createCommand(this.wiedervorlage$.value.resource, linkRel, commandProvider())
      .subscribe((commandStateResource) => {
        if (isDone(commandStateResource.resource)) {
          this.handleSource(commandStateResource.resource);
          this.showSnackBar(
            commandStateResource.resource,
            this.buildMessage(snackBarMessage, this.wiedervorlage$.value.resource),
          );
          commandStateSubj.next(commandStateResource);
          commandSubscription.unsubscribe();
        }
      });

    return commandStateSubj.asObservable();
  }

  private handleSource(commandResource: CommandResource): void {
    if (isDone(commandResource)) {
      this.loadAndSetWiedervorlageByUrl(this.getWiedervorlageUri());
    }
  }

  private buildMessage(message: string, wiedervorlage: Wiedervorlage) {
    return replacePlaceholder(message, 'betreff', wiedervorlage.betreff);
  }

  showSnackBar(commandResource: CommandResource, message: string) {
    if (isDone(commandResource)) {
      this.snackbarService.show(commandResource, message);
    }
  }

  public getAttachmentList(
    wiedervorlage: WiedervorlageResource,
  ): Observable<StateResource<BinaryFileListResource>> {
    const uri: ResourceUri = getUrl(wiedervorlage);

    if (isNil(this.attachmentListByWiedervorlage[uri])) {
      this.attachmentListByWiedervorlage[uri] = new BehaviorSubject(
        createEmptyStateResource<any>(),
      );
    } else {
      this.attachmentListByWiedervorlage[uri].next(createEmptyStateResource<any>());
    }

    doIfLoadingRequired(this.attachmentListByWiedervorlage[uri].value, () =>
      this.loadAttachments(wiedervorlage),
    );

    return this.attachmentListByWiedervorlage[uri].asObservable();
  }

  loadAttachments(wiedervorlage: WiedervorlageResource): void {
    this.setAttachmentLoading(wiedervorlage);
    if (!isNil(this.attachmentSubscription)) this.attachmentSubscription.unsubscribe();
    this.attachmentSubscription = this.binaryFileService
      .getFiles(wiedervorlage, WiedervorlageLinkRel.ATTACHMENTS)
      .subscribe((fileList) => {
        if (fileList.loaded) {
          this.attachmentListByWiedervorlage[getUrl(wiedervorlage)].next(fileList);
        }
      });
  }

  setAttachmentLoading(wiedervorlage: WiedervorlageResource): void {
    this.attachmentListByWiedervorlage[getUrl(wiedervorlage)].next({
      ...this.attachmentListByWiedervorlage[getUrl(wiedervorlage)].value,
      loading: true,
    });
  }

  public getSubmitInProgress(): Observable<StateResource<CommandResource>> {
    return this.submitInProgress$.asObservable();
  }

  ngOnDestroy(): void {
    this.unsubscribe();
  }

  private unsubscribe(): void {
    if (!isNil(this.subscription)) this.subscription.unsubscribe();
  }

  public hasEditLink(wiedervorlageUrl: string): Observable<boolean> {
    const decodedUrl: string = decodeUrlFromEmbedding(wiedervorlageUrl);

    return this.repository
      .getWiedervorlage(decodedUrl)
      .pipe(map((wiedervorlage) => hasLink(wiedervorlage, WiedervorlageLinkRel.EDIT)));
  }
}
