package de.ozgcloud.alfa.historie;

import static de.ozgcloud.alfa.historie.AktenzeichenChangeHistoryBuilder.*;
import static de.ozgcloud.alfa.historie.ChangeHistoryBuilderTest.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.LocalDateTime;
import java.util.List;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandTestFactory;

public class AktenzeichenChangeHistoryBuilderTest {

	private static final String ORGANISATIONSEINHEITEN_ID = UUID.randomUUID().toString();

	private final Command previousCommand = commandFinishedAt(LocalDateTime.of(2023, 5, 1, 12, 0));
	private final Command command = commandFinishedAt(LocalDateTime.of(2023, 6, 1, 12, 0));
	private final List<Command> commands = List.of(previousCommand, command);

	@Spy
	private AktenzeichenChangeHistoryBuilder builder = AktenzeichenChangeHistoryBuilder.builder()
			.withCommands(commands)
			.withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID);

	@Nested
	class TestIsRelevant {

		@Test
		void shouldReturnTrue() {
			var command = CommandTestFactory.createBuilder()
					.order(CommandOrder.SET_AKTENZEICHEN)
					.build();

			var result = builder.isRelevant(command);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalse() {
			var command = CommandTestFactory.createBuilder()
					.order(CommandOrder.VORGANG_WIEDEREROEFFNEN)
					.build();

			var result = builder.isRelevant(command);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestToCommandWithChangeValues {

		private static final String EXPECTED_AKTENZEICHEN_BEFORE_CHANGE = LoremIpsum.getInstance().getWords(2);
		private static final String EXPECTED_AKTENZEICHEN_AFTER_CHANGE = LoremIpsum.getInstance().getWords(2);

		private final CommandWithPrevious commandWithPrevious = CommandWithPreviousTestFactory.createBuilder().command(command).previous(previousCommand).build();

		@BeforeEach
		void init() {
			when(builder.getAktenzeichenBeforeChange(commandWithPrevious)).thenReturn(EXPECTED_AKTENZEICHEN_BEFORE_CHANGE);
			when(builder.getAktenzeichenAfterChange(commandWithPrevious)).thenReturn(EXPECTED_AKTENZEICHEN_AFTER_CHANGE);
		}

		@Test
		void shouldSetCommand() {
			var commandWithChangeValues = callBuilder();

			assertThat(commandWithChangeValues.command()).isEqualTo(command);
		}

		@Test
		void shouldSetValueBeforeChange() {
			var commandWithChangeValues = callBuilder();

			assertThat(commandWithChangeValues.valueBeforeChange()).isEqualTo(EXPECTED_AKTENZEICHEN_BEFORE_CHANGE);
		}

		@Test
		void shouldSetValueAfterChange() {
			var commandWithChangeValues = callBuilder();

			assertThat(commandWithChangeValues.valueAfterChange()).isEqualTo(EXPECTED_AKTENZEICHEN_AFTER_CHANGE);
		}

		private ChangeHistoryBuilder.CommandWithChangeValues callBuilder() {
			return builder.toCommandWithChangeValues(commandWithPrevious);
		}
	}

	@Nested
	class TestGetAktenzeichenBeforeChange {

		private final CommandWithPrevious commandWithoutPrevious = CommandWithPreviousTestFactory.createBuilder().command(previousCommand).previous(null).build();
		private final CommandWithPrevious commandWithPrevious = CommandWithPreviousTestFactory.createBuilder().command(command).previous(previousCommand).build();

		@Test
		void shouldReturnEmptyIfPreviousIsNull() {
			var aktenzeichen = builder.getAktenzeichenBeforeChange(commandWithoutPrevious);

			assertThat(aktenzeichen).isEmpty();
		}

		@Test
		void shouldGetAktenzeichenFromPreviousCommand() {
			var expectedValue = LoremIpsum.getInstance().getWords(2);
			when(builder.getAktenzeichen(previousCommand)).thenReturn(expectedValue);

			var aktenzeichen = builder.getAktenzeichenBeforeChange(commandWithPrevious);

			assertThat(aktenzeichen).isEqualTo(expectedValue);
		}
	}

	@Nested
	class TestGetAktenzeichenAfterChange {

		private static final String EXPECTED_AKTENZEICHEN = LoremIpsum.getInstance().getWords(2);

		private final CommandWithPrevious commandWithPrevious = CommandWithPreviousTestFactory.createBuilder().command(command).previous(previousCommand).build();

		@BeforeEach
		void init() {
			doReturn(EXPECTED_AKTENZEICHEN).when(builder).getAktenzeichen(command);
		}

		@Test
		void shouldGetAktenzeichenFromCommand() {
			builder.getAktenzeichenAfterChange(commandWithPrevious);

			verify(builder).getAktenzeichen(command);
		}

		@Test
		void shouldReturnAktenzeichen() {
			var result = builder.getAktenzeichenAfterChange(commandWithPrevious);

			assertThat(result).isEqualTo(EXPECTED_AKTENZEICHEN);
		}
	}

	@Nested
	class TestGetAktenzeichen {

		private static final String AKTENZEICHEN = LoremIpsum.getInstance().getWords(1);

		@BeforeEach
		void init() {
			when(builder.getValueFromCommandBody(BODY_PROPERTY_AKTENZEICHEN, previousCommand)).thenReturn(AKTENZEICHEN);
		}

		@Test
		void shouldGetValueFromCommandBody() {
			builder.getAktenzeichen(previousCommand);

			verify(builder).getValueFromCommandBody(BODY_PROPERTY_AKTENZEICHEN, previousCommand);
		}

		@Test
		void shouldReturnValueFromCommandBody() {
			var value = builder.getAktenzeichen(previousCommand);

			assertThat(value).isEqualTo(AKTENZEICHEN);
		}
	}
}
