package de.ozgcloud.alfa.historie;

import static de.ozgcloud.alfa.historie.ChangeHistoryBuilderTest.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.LocalDateTime;
import java.util.List;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandTestFactory;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;

public class StatusChangeHistoryBuilderTest {

	private static final String ORGANISATIONSEINHEITEN_ID = UUID.randomUUID().toString();

	private final Command previousCommand = commandFinishedAt(LocalDateTime.of(2023, 5, 1, 12, 0));
	private final Command command = commandFinishedAt(LocalDateTime.of(2023, 6, 1, 12, 0));
	private final List<Command> commands = List.of(previousCommand, command);

	@Spy
	private StatusChangeHistoryBuilder builder = StatusChangeHistoryBuilder.builder()
			.withCommands(commands)
			.withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID);

	@Nested
	class TestIsRelevant {

		@EnumSource(mode = Mode.INCLUDE, names = { "VORGANG_ANNEHMEN", "VORGANG_VERWERFEN", "VORGANG_ZURUECKHOLEN", "VORGANG_BEARBEITEN",
				"VORGANG_BESCHEIDEN", "VORGANG_ABSCHLIESSEN", "VORGANG_ZUM_LOESCHEN_MARKIEREN", "VORGANG_LOESCHEN", "VORGANG_ZURUECKSTELLEN",
				"VORGANG_WIEDEREROEFFNEN" })
		@ParameterizedTest
		void shouldReturnTrue(CommandOrder order) {
			var isStatusChangeCommand = builder.isRelevant(CommandTestFactory.createBuilder().order(order).build());

			assertThat(isStatusChangeCommand).isTrue();
		}

		@EnumSource(mode = Mode.EXCLUDE, names = { "VORGANG_ANNEHMEN", "VORGANG_VERWERFEN", "VORGANG_ZURUECKHOLEN", "VORGANG_BEARBEITEN",
				"VORGANG_BESCHEIDEN", "VORGANG_ABSCHLIESSEN", "VORGANG_ZUM_LOESCHEN_MARKIEREN", "VORGANG_LOESCHEN", "VORGANG_ZURUECKSTELLEN",
				"VORGANG_WIEDEREROEFFNEN" })
		@ParameterizedTest
		void shouldReturnFalse(CommandOrder order) {
			var isStatusChangeCommand = builder.isRelevant(CommandTestFactory.createBuilder().order(order).build());

			assertThat(isStatusChangeCommand).isFalse();
		}
	}

	@Nested
	class TestToCommandWithChangeValues {

		private static final String EXPECTED_STATUS_BEFORE_CHANGE = LoremIpsum.getInstance().getWords(2);
		private static final String EXPECTED_STATUS_AFTER_CHANGE = LoremIpsum.getInstance().getWords(2);

		private final CommandWithPrevious commandWithPrevious = CommandWithPreviousTestFactory.createBuilder().command(command)
				.previous(previousCommand).build();

		@BeforeEach
		void init() {
			doReturn(EXPECTED_STATUS_BEFORE_CHANGE).when(builder).getStatusBeforeChange(commandWithPrevious);
			doReturn(EXPECTED_STATUS_AFTER_CHANGE).when(builder).getStatusAfterChange(commandWithPrevious);
		}

		@Test
		void shouldSetCommand() {
			var commandWithChangeValues = callBuilder();

			assertThat(commandWithChangeValues.command()).isEqualTo(command);
		}

		@Test
		void shouldSetValueBeforeChange() {
			var commandWithChangeValues = callBuilder();

			assertThat(commandWithChangeValues.valueBeforeChange()).isEqualTo(EXPECTED_STATUS_BEFORE_CHANGE);
		}

		@Test
		void shouldSetValueAfterChange() {
			var commandWithChangeValues = callBuilder();

			assertThat(commandWithChangeValues.valueAfterChange()).isEqualTo(EXPECTED_STATUS_AFTER_CHANGE);
		}

		private ChangeHistoryBuilder.CommandWithChangeValues callBuilder() {
			return builder.toCommandWithChangeValues(commandWithPrevious);
		}
	}

	@Nested
	class TestGetStatusBeforeChange {

		private final CommandWithPrevious commandWithoutPrevious = CommandWithPreviousTestFactory.createBuilder().command(previousCommand)
				.previous(null).build();
		private final CommandWithPrevious commandWithPrevious = CommandWithPreviousTestFactory.createBuilder().command(command)
				.previous(previousCommand).build();

		@Test
		void shouldReturnStatusBeforeChange() {
			var status = "Status";
			doReturn(status).when(builder).getStatus(previousCommand);

			var statusBeforeChange = builder.getStatusBeforeChange(commandWithPrevious);

			assertThat(statusBeforeChange).isEqualTo(status);
		}

		@Test
		void shouldReturnDefaultValue() {
			var status = builder.getStatusBeforeChange(commandWithoutPrevious);

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.NEU));
		}
	}

	@Nested
	class TestGetStatusAfterChange {

		private final CommandWithPrevious commandWithPrevious = CommandWithPreviousTestFactory.createBuilder().command(command)
				.previous(previousCommand).build();

		@Test
		void shouldReturnStatusAfterChange() {
			var status = "Status";
			doReturn(status).when(builder).getStatus(command);

			var statusAfterChange = builder.getStatusAfterChange(commandWithPrevious);

			assertThat(statusAfterChange).isEqualTo(status);
		}
	}

	@Nested
	class TestGetStatus {

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "VORGANG_ANNEHMEN", "VORGANG_ZURUECKSTELLEN" })
		void shouldReturnAngenommen(CommandOrder order) {
			var status = builder.getStatus(CommandTestFactory.createBuilder().order(order).build());

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ANGENOMMEN));
		}

		@Test
		void shouldReturnVerworfen() {
			var status = builder.getStatus(CommandTestFactory.createBuilder().order(CommandOrder.VORGANG_VERWERFEN).build());

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.VERWORFEN));
		}

		@Test
		void shouldReturnNeu() {
			var status = builder.getStatus(CommandTestFactory.createBuilder().order(CommandOrder.VORGANG_ZURUECKHOLEN).build());

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.NEU));
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "VORGANG_BEARBEITEN", "VORGANG_WIEDEREROEFFNEN" })
		void shouldReturnInBearbeitung(CommandOrder order) {
			var status = builder.getStatus(CommandTestFactory.createBuilder().order(order).build());

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.IN_BEARBEITUNG));
		}

		@Test
		void shouldReturnBeschieden() {
			var status = builder.getStatus(CommandTestFactory.createBuilder().order(CommandOrder.VORGANG_BESCHEIDEN).build());

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.BESCHIEDEN));
		}

		@Test
		void shouldReturnAbschliessen() {
			var status = builder.getStatus(CommandTestFactory.createBuilder().order(CommandOrder.VORGANG_ABSCHLIESSEN).build());

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ABGESCHLOSSEN));
		}

		@Test
		void shouldReturnZuLoeschen() {
			var status = builder.getStatus(CommandTestFactory.createBuilder().order(CommandOrder.VORGANG_ZUM_LOESCHEN_MARKIEREN).build());

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.VORGANG_STATUS_TO_NAME.get(VorgangStatus.ZU_LOESCHEN));
		}

		@Test
		void shouldReturnGeloescht() {
			var status = builder.getStatus(CommandTestFactory.createBuilder().order(CommandOrder.VORGANG_LOESCHEN).build());

			assertThat(status).isEqualTo(StatusChangeHistoryBuilder.GELOESCHT_NAME);
		}

		@Test
		void shouldtThrowIllegalArgumentException() {
			assertThatThrownBy(() -> builder.getStatus(CommandTestFactory.createBuilder().order(CommandOrder.ASSIGN_USER).build())).isInstanceOf(
					IllegalArgumentException.class);
		}
	}
}
