/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.OutputStream;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.binaryfile.BinaryFileService;
import de.ozgcloud.alfa.common.binaryfile.BinaryFileTestFactory;
import de.ozgcloud.alfa.common.command.CommandService;
import de.ozgcloud.alfa.common.command.CommandTestFactory;
import de.ozgcloud.alfa.common.command.CreateCommand;
import de.ozgcloud.alfa.common.errorhandling.ResourceNotFoundException;
import de.ozgcloud.alfa.common.file.OzgFile;
import de.ozgcloud.alfa.common.file.OzgFileTestFactory;
import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserProfile;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserService;
import de.ozgcloud.alfa.vorgang.ServiceKontoTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangHeadTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;

class PostfachMailServiceTest {

	@Spy
	@InjectMocks // NOSONAR
	private PostfachMailService service;
	@Mock
	private PostfachNachrichtPdfService pdfService;
	@Mock
	private PostfachMailRemoteService remoteService;
	@Mock
	private BinaryFileService fileService;
	@Mock
	private UserService userService;
	@Mock
	private CommandService commandService;

	@DisplayName("Get all")
	@Nested
	class TestGetAll {

		@Test
		void shouldCallRemoteService() {
			service.getAll(VorgangHeaderTestFactory.ID);

			verify(remoteService).findPostfachMails(VorgangHeaderTestFactory.ID);
		}
	}

	@DisplayName("Find by id")
	@Nested
	class TestFindById {

		private final PostfachMail mail = PostfachMailTestFactory.create();

		@BeforeEach
		void init() {
			when(remoteService.findById(any())).thenReturn(Optional.of(mail));
		}

		@Test
		void shouldCallRemoteService() {
			service.findById(PostfachMailTestFactory.NACHRICHT_ID);

			verify(remoteService).findById(PostfachMailTestFactory.NACHRICHT_ID);
		}

		@Test
		void shouldThrowExceptionIfNotFound() {
			when(remoteService.findById(any())).thenReturn(Optional.empty());

			assertThrows(ResourceNotFoundException.class, () -> service.findById(PostfachMailTestFactory.NACHRICHT_ID));
		}

		@Test
		void shouldReturnNachricht() {
			var result = service.findById(PostfachMailTestFactory.NACHRICHT_ID);

			assertThat(result).isSameAs(mail);
		}
	}

	@DisplayName("Send postfach mail")
	@Nested
	class TestSendPostfachMail {

		private static final CreateCommand CREATE_COMMAND = CommandTestFactory.createCreateCommand();

		@Test
		void shouldCallCommandService() {
			service.sendPostfachMail(CREATE_COMMAND, VorgangHeaderTestFactory.VERSION);

			verify(commandService).createCommand(CREATE_COMMAND, VorgangHeaderTestFactory.VERSION);
		}
	}

	@DisplayName("Resend postfach mail")
	@Nested
	class TestResendPostfachMail {

		@Test
		void shouldCallRemoteService() {
			service.resendPostfachMail(CommandTestFactory.ID, PostfachMailTestFactory.ID);

			verify(remoteService).resendPostfachMail(CommandTestFactory.ID, PostfachMailTestFactory.ID);
		}
	}

	@DisplayName("Is postfach configured")
	@Nested
	class TestIsPostfachConfigured {

		@Test
		void shouldCallRemoteServiceOnMissingProperty() {
			when(remoteService.getPostfachConfig()).thenReturn(PostfachConfigGroupTestFactory.create());

			service.isPostfachConfigured();

			verify(remoteService).getPostfachConfig();
		}

		@Test
		void shouldNotCallRemoteServiceOnExistingProperty() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.create());

			service.isPostfachConfigured();

			verifyNoInteractions(remoteService);
		}

		@Test
		void shoulCallRemoteServiceOnlyOnce() {
			when(remoteService.getPostfachConfig()).thenReturn(PostfachConfigGroupTestFactory.create());

			service.isPostfachConfigured();
			service.isPostfachConfigured();

			verify(remoteService, times(1)).getPostfachConfig();
		}

		@ParameterizedTest
		@ValueSource(booleans = { false, true })
		void shouldReturnValueFromRemoteService(boolean expectedConfigured) {
			var serviceResponse = PostfachConfigGroupTestFactory.createBuilder().configured(expectedConfigured).build();
			when(remoteService.getPostfachConfig()).thenReturn(serviceResponse);

			var configured = service.isPostfachConfigured();

			assertThat(configured).isEqualTo(expectedConfigured);
		}

		@ParameterizedTest
		@ValueSource(booleans = { false, true })
		void shouldReturnCachedValue(boolean expectedConfigured) {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.createBuilder().configured(expectedConfigured).build());

			var configured = service.isPostfachConfigured();

			assertThat(configured).isEqualTo(expectedConfigured);
		}
	}

	@DisplayName("Get all as pdf")
	@Nested
	class TestGetAllAsPdf {

		@Mock
		private OutputStream outputStream;

		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		private final Stream<PostfachNachrichtPdfData> postfachNachrichtPdfModelDataList = Stream
				.of(PostfachNachrichtPdfData.builder().build());// TODO TestFactory erstellen

		@DisplayName("should call")
		@Nested
		class TestCall {

			@BeforeEach
			void mock() {
				doReturn(postfachNachrichtPdfModelDataList).when(service).buildPostfachNachrichtPdfDataList(anyString());
			}

			@Test
			void buildPostfachNachrichtPdfModelDataList() {
				service.getAllAsPdf(vorgang, outputStream);

				verify(service).buildPostfachNachrichtPdfDataList(VorgangHeaderTestFactory.ID);
			}

			@Test
			void pdfService() {
				service.getAllAsPdf(vorgang, outputStream);

				verify(pdfService).getAllAsPdf(vorgang, postfachNachrichtPdfModelDataList, outputStream);
			}
		}

		@DisplayName("build postfach nachrichten pdf data")
		@Nested
		class TestBuildPostfachNachrichtPdfDataList {

			private final PostfachMail postfachNachricht = PostfachMailTestFactory.create();

			@DisplayName("should call")
			@Nested
			class TestCall {

				private final Stream<PostfachMail> postfachMail = Stream.of(postfachNachricht);
				private final Stream<OzgFile> ozgFile = Stream.of(OzgFileTestFactory.create());

				@BeforeEach
				void mock() {
					when(remoteService.findPostfachMails(anyString())).thenReturn(postfachMail);
					when(fileService.getFiles(anyList())).thenReturn(ozgFile);
				}

				@Test
				void shouldCallRemoteService() {
					service.buildPostfachNachrichtPdfDataList(VorgangHeaderTestFactory.ID);

					verify(remoteService).findPostfachMails(VorgangHeaderTestFactory.ID);
				}

				@Test
				void shouldCallBinaryFileService() {
					service.buildPostfachNachrichtPdfDataList(VorgangHeaderTestFactory.ID);

					verify(fileService).getFiles(List.of(BinaryFileTestFactory.FILE_ID));
				}

				@Test
				void shouldCallBuildPostfachNachrichtPdfData() {
					doReturn(PostfachNachrichtPdfData.builder().build()).when(service).buildPostfachNachrichtPdfData(any(), any());

					service.buildPostfachNachrichtPdfDataList(VorgangHeaderTestFactory.ID).toList();

					verify(service).buildPostfachNachrichtPdfData(postfachNachricht, Map.of(OzgFileTestFactory.ID, OzgFileTestFactory.NAME));
				}
			}

			@DisplayName("for single postfachNachricht")
			@Nested
			class TestBuildPostfachNachrichtPdfData {

				private final UserProfile user = UserProfileTestFactory.create();

				@BeforeEach
				void mock() {
					when(userService.getById(any(UserId.class))).thenReturn(user);
				}

				@Test
				void shouldCallUserService() {
					buildPostfachNachrichtPdfData();

					verify(userService).getById(UserProfileTestFactory.ID);
				}

				@Test
				void shouldHaveSetCreatedAt() {
					var postfachNachricht = buildPostfachNachrichtPdfData();

					assertThat(postfachNachricht.getCreatedAt()).isEqualTo(PostfachMailTestFactory.CREATED_AT);
				}

				@Test
				void shouldHaveSetCreatedByName() {
					var postfachNachricht = buildPostfachNachrichtPdfData();

					assertThat(postfachNachricht.getUser()).isEqualTo(user);
				}

				@Test
				void shouldHaveSetSubject() {
					var postfachNachricht = buildPostfachNachrichtPdfData();

					assertThat(postfachNachricht.getSubject()).isEqualTo(PostfachMailTestFactory.SUBJECT);
				}

				@Test
				void shouldHaveSetMailBody() {
					var postfachNachricht = buildPostfachNachrichtPdfData();

					assertThat(postfachNachricht.getMailBody()).isEqualTo(PostfachMailTestFactory.MAIL_BODY);
				}

				@Test
				void shouldHaveSetAttachmentNames() {
					var postfachNachricht = buildPostfachNachrichtPdfData();

					assertThat(postfachNachricht.getAttachmentNames()).isEqualTo(List.of(OzgFileTestFactory.NAME));
				}
			}

			@DisplayName("user")
			@Nested
			class TestUser {

				@DisplayName("exists")
				@Nested
				class TestOnNonNull {

					private final UserProfile user = UserProfileTestFactory.create();

					@BeforeEach
					void mock() {
						when(userService.getById(any(UserId.class))).thenReturn(user);
					}

					@Test
					void shouldCallUserService() {
						buildPostfachNachrichtPdfData();

						verify(userService).getById(UserProfileTestFactory.ID);
					}

					@Test
					void shouldHaveSetCreatedByName() {
						var postfachNachricht = buildPostfachNachrichtPdfData();

						assertThat(postfachNachricht.getUser()).isEqualTo(user);
					}
				}

				@DisplayName("not exists")
				@Nested
				class TestOnNull {

					private final PostfachMail postfachNachrichtWithoutCreatedBy = PostfachMailTestFactory.createBuilder().createdBy(null)
							.build();

					@Test
					void shouldNotCallUserService() {
						buildPostfachNachrichtPdfDataWithoutUser();

						verify(userService, never()).getById(any());
					}

					@Test
					void shouldHaveSetNullAsUser() {
						var postfachNachricht = buildPostfachNachrichtPdfDataWithoutUser();

						assertThat(postfachNachricht.getUser()).isNull();
					}

					private PostfachNachrichtPdfData buildPostfachNachrichtPdfDataWithoutUser() {
						return service.buildPostfachNachrichtPdfData(postfachNachrichtWithoutCreatedBy,
								Map.of(BinaryFileTestFactory.FILE_ID, OzgFileTestFactory.NAME));
					}
				}
			}

			private PostfachNachrichtPdfData buildPostfachNachrichtPdfData() {
				return service.buildPostfachNachrichtPdfData(postfachNachricht, Map.of(BinaryFileTestFactory.FILE_ID, OzgFileTestFactory.NAME));
			}
		}

		@DisplayName("get file ids from all attachments")
		@Nested
		class TestGetFileIdsFromAllAttachments {

			@Test
			void shouldReturnFileIdsFromPostfachMail() {
				var fileIds = service.getFileIdsFromAllAttachments(Stream.of(PostfachMailTestFactory.create()));

				assertThat(fileIds).containsExactly(BinaryFileTestFactory.FILE_ID);
			}

			@Test
			void shouldReturnEmptyListOnNoAttachments() {
				var fileIds = service.getFileIdsFromAllAttachments(Stream.of(PostfachMailTestFactory.createBuilder().clearAttachments().build()));

				assertThat(fileIds).isEmpty();
			}
		}

		@DisplayName("get User")
		@Nested
		class TestGetUser {

			@DisplayName("on existing client user")
			@Nested
			class TestOnExistingClientUser {

				@Test
				void shouldCallUserServiceOnExistingUser() {
					service.getUser(PostfachMailTestFactory.create());

					verify(userService).getById(UserProfileTestFactory.ID);
				}

				@Test
				void shouldReturnUserProfileWithId() {
					when(userService.getById(any())).thenReturn(UserProfileTestFactory.create());

					var user = service.getUser(PostfachMailTestFactory.create());

					assertThat(user.getId()).isEqualTo(UserProfileTestFactory.ID);
				}
			}

			@DisplayName("on existing system user")
			@Nested
			class TestOnSystemUser {

				private final PostfachMail postfachMail = PostfachMailTestFactory.createBuilder()
						.createdBy(UserId.from(UserProfileTestFactory.SYSTEM_USER)).build();

				@Test
				void shouldNotCallUserService() {
					service.getUser(postfachMail);

					verify(userService, never()).getById(UserProfileTestFactory.ID);
				}

				@Test
				void shouldReturnUserProfileWithId() {
					var user = service.getUser(postfachMail);

					assertThat(user.getId()).hasToString(UserProfileTestFactory.SYSTEM_USER);
				}
			}
		}
	}

	@Nested
	@DisplayName("Is reply to a message allowed for given Vorgang")
	class TestIsReplyToMessageAllowed {

		@Test
		void shouldCallIsReplyAllowed() {
			doReturn(true).when(service).isReplyAllowed(any());

			service.isReplyToMessageAllowed(VorgangWithEingangTestFactory.create());

			verify(service).isReplyAllowed(ServiceKontoTestFactory.TYPE);
		}

		@ParameterizedTest
		@ValueSource(booleans = { false, true })
		void shouldReturnResultOfIsReplyAllowed(boolean expectedReplyAllowed) {
			doReturn(expectedReplyAllowed).when(service).isReplyAllowed(any());

			var replyAllowed = service.isReplyToMessageAllowed(VorgangWithEingangTestFactory.create());

			assertThat(replyAllowed).isEqualTo(expectedReplyAllowed);
		}

		@Test
		void shouldReturnFalseForNotExistingServiceKonto() {
			var vorgangWithEingang = VorgangWithEingangTestFactory.createBuilder()
					.header(VorgangHeadTestFactory.createBuilder()
							.serviceKonto(null)
							.build())
					.build();

			var replyToMessageAllowed = service.isReplyToMessageAllowed(vorgangWithEingang);

			assertThat(replyToMessageAllowed).isFalse();
		}
	}

	@Nested
	class TestIsReplyAllowed {

		private static final String DUMMY_SERVICE_KONTO_TYPE = LoremIpsum.getInstance().getWords(1);

		@Test
		@DisplayName("reply not allowed for given ServiceKonto-type")
		void shouldReturnFalseIfPostfachNotConfigured() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.createBuilder().postfachConfigs(List.of()).build());
			doReturn(false).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(PostfachConfigTestFactory.TYPE);

			assertThat(replyAllowed).isFalse();
		}

		@Test
		@DisplayName("given ServiceKonto-type is not configured")
		void shouldReturnFalseForNotConfiguredType() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.createBuilder().postfachConfigs(List.of()).build());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(DUMMY_SERVICE_KONTO_TYPE);

			assertThat(replyAllowed).isFalse();
		}

		@Test
		@DisplayName("given ServiceKonto-type is null")
		void shouldReturnFalseForUnknownType() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.create());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(null);

			assertThat(replyAllowed).isFalse();
		}

		@Test
		@DisplayName("reply not allowed for given ServiceKonto-type")
		void shouldReturnFalseIfReplyNotAllowed() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.createBuilder()
					.postfachConfigs(List.of(PostfachConfigTestFactory.createBuilder().replyAllowed(false).build()))
					.build());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(PostfachConfigTestFactory.TYPE);

			assertThat(replyAllowed).isFalse();
		}

		@Test
		@DisplayName("reply allowed for given ServiceKonto-type")
		void shouldReturnTrueIfReplyIsAllowed() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.create());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(PostfachConfigTestFactory.TYPE);

			assertThat(replyAllowed).isTrue();
		}
	}

	private void setPostfachConfigGroup(PostfachConfigGroup postfachConfigGroup) {
		ReflectionTestUtils.setField(service, "postfachConfigGroup", postfachConfigGroup);
	}
}