package de.ozgcloud.alfa.kommentar;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;

@NoArgsConstructor(access = AccessLevel.PRIVATE)
public final class DateiformatCode {

	static final String SONSTIGES_CODE = "100";

	// MIME type -> file extension -> xdomea code
	private static final Map<String, Map<String, String>> mimeTypesMapping = new HashMap<>();

	static {
		mimeTypesMapping.put("image/bmp", Map.of("bmp", "001"));
		mimeTypesMapping.put("text/csv", Map.of("csv", "002"));
		mimeTypesMapping.put("application/msword", Map.of("doc", "003", "dot", "032"));
		mimeTypesMapping.put("message/rfc822", Map.of("eml", "006"));
		mimeTypesMapping.put("text/html", Map.of("htm", "007", "html", "008"));
		mimeTypesMapping.put("image/jpeg", Map.of("jpe", "009", "jpeg", "010", "jpg", "011"));
		mimeTypesMapping.put("application/vnd.ms-outlook", Map.of("msg", "012"));
		mimeTypesMapping.put("application/vnd.oasis.opendocument.presentation", Map.of("odp", "013"));
		mimeTypesMapping.put("application/vnd.oasis.opendocument.spreadsheet", Map.of("ods", "014"));
		mimeTypesMapping.put("application/vnd.oasis.opendocument.text", Map.of("odt", "015"));
		mimeTypesMapping.put("application/vnd.oasis.opendocument.presentation-template", Map.of("otp", "016"));
		mimeTypesMapping.put("application/vnd.oasis.opendocument.spreadsheet-template", Map.of("ots", "017"));
		mimeTypesMapping.put("application/pdf", Map.of("pdf", "018"));
		mimeTypesMapping.put("image/png", Map.of("png", "019"));
		mimeTypesMapping.put("application/vnd.ms-powerpoint", Map.of("ppt", "020", "pot", "033"));
		mimeTypesMapping.put("application/postscript", Map.of("ps", "021"));
		mimeTypesMapping.put("application/rtf", Map.of("rtf", "022"));
		mimeTypesMapping.put("image/tiff", Map.of("tif", "023", "tiff", "024"));
		mimeTypesMapping.put("application/vnd.ms-works", Map.of("wps", "025"));
		mimeTypesMapping.put("application/vnd.ms-excel", Map.of("xlc", "026", "xlm", "027", "xls", "028", "xlw", "029", "xlt", "034"));
		mimeTypesMapping.put("application/xml", Map.of("xml", "030", "xsd", "031"));
		mimeTypesMapping.put("text/xml", Map.of("xml", "030", "xsd", "031"));
		mimeTypesMapping.put("application/vnd.openxmlformats-officedocument.wordprocessingml.document", Map.of("docx", "035"));
		mimeTypesMapping.put("application/vnd.openxmlformats-officedocument.spreadsheetml.sheet", Map.of("xlsx", "036"));
		mimeTypesMapping.put("application/vnd.openxmlformats-officedocument.presentationml.presentation", Map.of("pptx", "037"));
		mimeTypesMapping.put("application/vnd.openxmlformats-officedocument.wordprocessingml.template", Map.of("dotx", "038"));
		mimeTypesMapping.put("application/vnd.openxmlformats-officedocument.spreadsheetml.template", Map.of("xltx", "039"));
		mimeTypesMapping.put("application/vnd.openxmlformats-officedocument.presentationml.template", Map.of("potx", "040"));
		mimeTypesMapping.put("application/xml-dtd", Map.of("dtd", "041"));
		mimeTypesMapping.put("application/vnd.ms-project", Map.of("mpp", "042"));
		mimeTypesMapping.put("application/vnd.oasis.opendocument.text-template", Map.of("ott", "043"));
		mimeTypesMapping.put("text/plain", Map.of("txt", "045"));
		mimeTypesMapping.put("application/x-pkcs7-certificates", Map.of("p7b", "046"));
		mimeTypesMapping.put("application/pkcs7-mime", Map.of("p7c", "047", "p7m", "048"));
		mimeTypesMapping.put("application/pkcs7-signature", Map.of("p7s", "049"));
		mimeTypesMapping.put("application/vnd.etsi.asic-s+zip", Map.of("asics", "050"));
		mimeTypesMapping.put("application/scvp-cv-response", Map.of("scs", "051"));
	}

	public static String getXdomeaCode(String mimeType, String extension) {
		var fileExtensionToCodeMapping = getFileExtensionToCodeMapping(mimeType);
		if (hasSingleExtensionMapping(fileExtensionToCodeMapping)) {
			return getCodeOrDefault(fileExtensionToCodeMapping);
		}
		return getCodeForExtensionOrDefault(extension, fileExtensionToCodeMapping);
	}

	private static Map<String, String> getFileExtensionToCodeMapping(String mimeType) {
		return mimeTypesMapping.getOrDefault(Objects.requireNonNullElse(mimeType, StringUtils.EMPTY), Collections.emptyMap());
	}

	private static boolean hasSingleExtensionMapping(Map<String, String> fileExtensionToCodeMapping) {
		return fileExtensionToCodeMapping.entrySet().size() == 1;
	}

	private static String getCodeOrDefault(Map<String, String> fileExtensionToCodeMapping) {
		return fileExtensionToCodeMapping.values().stream().findFirst().orElse(SONSTIGES_CODE);
	}

	private static String getCodeForExtensionOrDefault(String extension, Map<String, String> fileExtensionToCodeMapping) {
		return fileExtensionToCodeMapping.getOrDefault(Objects.requireNonNullElse(extension, StringUtils.EMPTY), SONSTIGES_CODE);
	}

}
