package de.ozgcloud.alfa.file;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.OutputStream;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.springframework.http.MediaType;

import de.ozgcloud.alfa.common.binaryfile.BinaryFileService;
import de.ozgcloud.alfa.common.binaryfile.FileId;
import de.ozgcloud.alfa.common.file.OzgFile;
import de.ozgcloud.alfa.common.file.OzgFileService;
import de.ozgcloud.alfa.common.file.OzgFileTestFactory;
import de.ozgcloud.alfa.export.DokumentTypeTestFactory;
import de.ozgcloud.alfa.vorgang.EingangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.EingangTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import de.xoev.xdomea.DokumentType;

class ExportFileServiceTest {

	@InjectMocks
	private ExportFileService service;

	@Mock
	private OzgFileService ozgFileService;

	@Mock
	private BinaryFileService binaryFileService;

	@Nested
	class TestGetRepresentations {

		private final OzgFile representationPdfFile = OzgFileTestFactory.createBuilder().contentType(MediaType.APPLICATION_PDF_VALUE).build();
		private final OzgFile representationXmlFile = OzgFileTestFactory.createBuilder().contentType(MediaType.TEXT_XML_VALUE).build();

		@Nested
		class OnAvailableVorgangData {

			@BeforeEach
			void init() {
				when(ozgFileService.getRepresentationsByEingang(EingangTestFactory.ID)).thenReturn(
						Stream.of(representationPdfFile, representationXmlFile));
			}

			@Test
			void shouldLoadRepresentations() {
				service.getRepresentations(VorgangWithEingangTestFactory.create()).toList();

				verify(ozgFileService).getRepresentationsByEingang(EingangTestFactory.ID);
			}

			@Test
			void shouldReturnPdfs() {
				var representations = service.getRepresentations(VorgangWithEingangTestFactory.create());

				assertThat(representations).hasSize(1).containsExactlyInAnyOrder(representationPdfFile);
			}
		}

		@Nested
		class OnMissingVorgangData {

			@Test
			void shouldReturnEmptyOnMissingEingang() {
				var vorgang = VorgangWithEingangTestFactory.createBuilder().eingang(null).build();

				var representations = service.getRepresentations(vorgang);

				assertThat(representations).isEmpty();
			}

			@Test
			void shouldReturnEmptyOnMissingEingangId() {
				var vorgang = VorgangWithEingangTestFactory.createBuilder().eingang(EingangTestFactory.createBuilder().id(null).build()).build();

				var representations = service.getRepresentations(vorgang);

				assertThat(representations).isEmpty();
			}

		}

	}

	@Nested
	class TestGetAttachments {

		private final OzgFile attachmentPdfFile = OzgFileTestFactory.createBuilder().contentType(MediaType.APPLICATION_PDF_VALUE).build();
		private final OzgFile attachmentXmlFile = OzgFileTestFactory.createBuilder().contentType(MediaType.TEXT_XML_VALUE).build();

		@Nested
		class OnAvailableVorgangData {

			@BeforeEach
			void init() {
				when(ozgFileService.getAttachmentsByEingang(EingangTestFactory.ID)).thenReturn(Stream.of(attachmentXmlFile, attachmentPdfFile));
			}

			@Test
			void shouldLoadAttachments() {
				service.getAttachments(VorgangWithEingangTestFactory.create()).toList();

				verify(ozgFileService).getAttachmentsByEingang(EingangTestFactory.ID);
			}

			@Test
			void shouldReturnPdfs() {
				var ozgFiles = service.getAttachments(VorgangWithEingangTestFactory.create());

				assertThat(ozgFiles).hasSize(1).containsExactlyInAnyOrder(attachmentPdfFile);
			}
		}

		@Nested
		class OnMissingVorgangData {
			@Test
			void shouldReturnEmptyOnMissingEingang() {
				var vorgang = VorgangWithEingangTestFactory.createBuilder().eingang(null).build();

				var attachments = service.getAttachments(vorgang);

				assertThat(attachments).isEmpty();
			}

			@Test
			void shouldReturnEmptyOnMissingEingangId() {
				var vorgang = VorgangWithEingangTestFactory.createBuilder().eingang(EingangTestFactory.createBuilder().id(null).build()).build();

				var attachments = service.getAttachments(vorgang);

				assertThat(attachments).isEmpty();
			}
		}
	}

	@Nested
	class TestWriteOzgFile {

		private final FileId fileId = OzgFileTestFactory.ID;

		@Mock
		private OutputStream outputStream;

		@Test
		void shouldCallBinaryFileService() {
			service.writeOzgFile(fileId, outputStream);

			verify(binaryFileService).writeFileContent(fileId, outputStream);
		}
	}

	@Nested
	class TestCreateDokumentType {

		private MockedStatic<DokumentTypeBuilder> dokumentTypeBuilderMockedStatic;

		@Mock
		private DokumentTypeBuilder dokumentTypeBuilder;

		private final OzgFile ozgFile = OzgFileTestFactory.create();
		private final DokumentType dokumentType = DokumentTypeTestFactory.create();

		@BeforeEach
		void setUp() {
			dokumentTypeBuilderMockedStatic = mockStatic(DokumentTypeBuilder.class);
			dokumentTypeBuilderMockedStatic.when(DokumentTypeBuilder::builder).thenReturn(dokumentTypeBuilder);

			when(dokumentTypeBuilder.withOzgFile(ozgFile)).thenReturn(dokumentTypeBuilder);
			when(dokumentTypeBuilder.withOrdinalNumber(1L)).thenReturn(dokumentTypeBuilder);
			when(dokumentTypeBuilder.withFormEngineName(EingangHeaderTestFactory.FORM_ENGINE_NAME)).thenReturn(dokumentTypeBuilder);
			when(dokumentTypeBuilder.build()).thenReturn(dokumentType);
		}

		@AfterEach
		void tearDown() {
			dokumentTypeBuilderMockedStatic.close();
		}

		@Test
		void shouldBuild() {
			service.createDokumentTypes(List.of(ozgFile), EingangHeaderTestFactory.FORM_ENGINE_NAME).toList();

			verify(dokumentTypeBuilder).build();
		}

		@Test
		void shouldSetOzgFile() {
			service.createDokumentTypes(List.of(ozgFile), EingangHeaderTestFactory.FORM_ENGINE_NAME).toList();

			verify(dokumentTypeBuilder).withOzgFile(ozgFile);
		}

		@Test
		void shouldSetOrdinalNumber() {
			service.createDokumentTypes(List.of(ozgFile), EingangHeaderTestFactory.FORM_ENGINE_NAME).toList();

			verify(dokumentTypeBuilder).withOrdinalNumber(1L);
		}

		@Test
		void shouldSetFormEngineName() {
			service.createDokumentTypes(List.of(ozgFile), EingangHeaderTestFactory.FORM_ENGINE_NAME).toList();

			verify(dokumentTypeBuilder).withFormEngineName(EingangHeaderTestFactory.FORM_ENGINE_NAME);
		}

		@Test
		void shouldReturnDokumentTypes() {
			var dokumentTypes = service.createDokumentTypes(List.of(ozgFile), EingangHeaderTestFactory.FORM_ENGINE_NAME);

			assertThat(dokumentTypes).containsExactly(dokumentType);
		}
	}

}